module Hit ( parseHit, Hit (Hit, ip, time, query, status, agent), 
	     epoch, future, testHit,
	     HTTPStatus (Ok, Error),
	     Agent (Bot, User, Unknown),
	     Query (Browse, Search, Getfile, Export, Other)
	   ) where

import HUnit
import Text.Regex
import System.Time
import Data.FiniteMap
import Monad


type IP = String

data HTTPStatus = Ok | Error Int deriving (Show, Eq)

data Agent = Bot | User | Unknown deriving (Show, Eq)

type URL = String

data Query = Browse URL | Search URL | Getfile URL | Export URL | Other URL deriving (Show, Eq)

data Hit = Hit { ip     :: IP,
		 time   :: ClockTime,
		 query  :: Query,
		 agent  :: Agent,
		 status :: HTTPStatus
	       } deriving Eq

instance Show Hit where
    show (Hit ip time query agent status) 
        = "Hit {" ++ (show ip) 
                  ++ " / " ++ (calendarTimeToString $ toUTCTime $ time) 
                  ++ " / " ++ (show query) 
                  ++ " / " ++ (show agent) 
                  ++ " / " ++ (show status)
                  ++ "}"

-- check if the user agent is a bot or not

_mkAgent name 
    = let _mkAgent' Nothing  = User
          _mkAgent' (Just _) = Bot

      in _mkAgent' (matchRegex _isBot name)
    where _isBot = mkRegex ".*(msnbot|Googlebot|Python-urllib|Inktomi)"


_testAgent = TestCase ( do assertEqual "MSN"    Bot (_mkAgent "msnbot/0.3")
                           assertEqual "Google" Bot (_mkAgent "Googlebot/2.1")
                      )


-- categorize the different queries

_mkQuery url
    = let _head "/" args           = Browse args
          _head "/search.py"  args = Search args
          _head "/getfile.py" args = Getfile args
          _head "/export.py" args  = Export args
          _head _ _                = Other url

          _check (Just [root, args]) = _head root args
          _check _                   = Other url

      in _check (matchRegex _rxQuery url)

    where _rxQuery = mkRegex "GET (/[^? ]*)(\\?[^ ]*)?"


_testQuery = TestCase 
             ( do assertEqual "index" 
                                  (Browse "") 
                                  (_mkQuery "GET /")
                  assertEqual "with args"
                                  (Browse "?ln=fr") 
                                  (_mkQuery "GET /?ln=fr")
                  assertEqual "no posts"
                                  (Other "POST /") 
                                  (_mkQuery "POST /")
                  assertEqual "no other links"
                                  (Other "GET /submit.py") 
                                  (_mkQuery "GET /submit.py")
                  assertEqual "with http version"
                                  (Browse "") 
                                  (_mkQuery "GET / HTTP/1.1")
                  assertEqual "search"
                                  (Search "?f=author&p=gobry")
                                  (_mkQuery "GET /search.py?f=author&p=gobry")
             )


-- Handling of log timestamps
_mkMonth month 
    = case lookupFM _months month of
      Nothing -> error ("invalid month " ++ month)
      Just m  -> m
    where _months = listToFM [ ("Jan", January),
			       ("Feb", February),
			       ("Mar", March),
			       ("Apr", April),
			       ("May", May),
			       ("Jun", June),
			       ("Jul", July),
			       ("Aug", August),
			       ("Sep", September),
			       ("Oct", October),
			       ("Nov", November),
			       ("Dec", December) ]

_mkTZ (sign:tz)
    = let (q, r) = divMod (read tz :: Int) 100
	  s      = if (sign == '-') then -1 else 1
      in s * (60 * q + r) * 60

_mkTZ "" = error "empty TZ"


_timeSplit :: String -> (Int, Month, Int, Int, Int, Int, Int)

_timeSplit t = ( read year :: Int,
		 _mkMonth month,
		 read day :: Int,
		 read h :: Int,
		 read m :: Int,
		 read s :: Int,
		 _mkTZ tz
	       )
    where [day, month, year, h, m, s, tz] 
	      = case matchRegex _timeRe t of
		Nothing    -> error ("invalid time: " ++ t)
		Just parts -> parts
	  _timeRe = mkRegex "([0-9]+)/([a-zA-Z]+)/([0-9]+):([0-9]+):([0-9]+):([0-9]+) (.[0-9]+)"



-- Parse a log timestamp into a calendar time
timeParse :: String -> ClockTime

timeParse t 
    = let (year, month, day, hour, mins, sec, tz) = _timeSplit t
	  in toClockTime (CalendarTime { ctYear  = year,
					 ctMonth = month,
					 ctDay   = day,
					 ctHour  = hour,
					 ctMin   = mins,
					 ctSec   = sec,
					 ctTZ    = tz,
					 ctPicosec = 0,
					 ctWDay = undefined, 
					 ctYDay = undefined, 
					 ctTZName = undefined, 
					 ctIsDST = False
				       })

-- Convenience "minimal value" for the time stamps
epoch :: ClockTime

epoch = toClockTime (CalendarTime { ctYear  = 1970,
				    ctMonth = January,
				    ctDay   = 1,
				    ctHour  = 0,
				    ctMin   = 0,
				    ctSec   = 0,
				    ctTZ    = 0,
				    ctPicosec = 0,
				    ctWDay = undefined, 
				    ctYDay = undefined, 
				    ctTZName = undefined, 
				    ctIsDST = False
				  })

future :: ClockTime

future = toClockTime (CalendarTime { ctYear  = 2020,
				     ctMonth = January,
				     ctDay   = 1,
				     ctHour  = 0,
				     ctMin   = 0,
				     ctSec   = 0,
				     ctTZ    = 0,
				     ctPicosec = 0,
				     ctWDay = undefined, 
				     ctYDay = undefined, 
				     ctTZName = undefined, 
				     ctIsDST = False
				   })



_short = "^([a-z0-9.-]+) - - \\[([^]]+)\\] \"([^\"]+)\" ([0-9]+) ([0-9]+|-)"

_lineRe  = mkRegex (_short ++ " \"([^\"]*)\" \"([^\"]*)\"$")
_shortRe = mkRegex (_short ++ "$")

_mkStatus status 
    = let v = read status :: Int
      in if (v == 200) then Ok else Error v
      

parseHit :: String -> Maybe Hit

parseHit line
    = let short line 
              = do [client, stamp, query, status, _] <- matchRegex _shortRe line
                   return (Hit client (timeParse stamp) (_mkQuery query) Unknown (_mkStatus status))
          long line
              = do [client, stamp, query, status, _, _, agent] <- matchRegex _lineRe line
                   return (Hit client (timeParse stamp) (_mkQuery query) (_mkAgent agent) (_mkStatus status))
      in (long line) `mplus` (short line)


_testParseHit 
    = TestCase 
      ( do assertEqual "long log" 
                       (Just (Hit "127.0.0.1" time (Browse "") User Ok))
                       (parseHit "127.0.0.1 - - [18/Jan/2005:13:00:38 +0100] \"GET / HTTP/1.1\" 200 309 \"-\" \"Mozilla/5.0\"")
           assertEqual "short log"
                       (Just (Hit "127.0.0.1" time (Browse "") Unknown Ok))
                       (parseHit "127.0.0.1 - - [18/Jan/2005:13:00:38 +0100] \"GET / HTTP/1.1\" 200 309")
           assertEqual "not found"
                       (Just (Hit "127.0.0.1" time (Browse "") User (Error 404)))
                       (parseHit "127.0.0.1 - - [18/Jan/2005:13:00:38 +0100] \"GET / HTTP/1.1\" 404 309 \"-\" \"Mozilla/5.0\"")
      )
    where time = toClockTime (CalendarTime { ctYear  = 2005, ctMonth = January, ctDay   = 18,
				             ctHour  = 13,   ctMin   = 0,       ctSec   = 38,
				             ctTZ    = 3600, ctPicosec = 0,
				             ctWDay = undefined, ctYDay = undefined, 
				             ctTZName = undefined, ctIsDST = False
				           })


testHit = runTestTT (TestList [_testAgent, _testQuery, _testParseHit])
