module ParseLog (parseLog, testParseLog) where

import Hit
import System.Time
import Control.Monad.State
import qualified HUnit 

data Stat = Stat { firstHit   :: !ClockTime,
		   lastHit    :: !ClockTime, 
		   total      :: !Int,
		   success    :: !Int,
		   user       :: !Int,
		   searches   :: !Int,
		   exports    :: !Int,
		   getfiles   :: !Int,
                   invalid    :: !Int
		 } deriving (Show, Eq)

initialStat = Stat { firstHit   = future,
		     lastHit    = epoch,
		     total      = 0,
		     success    = 0,
		     user       = 0,
		     searches   = 0,
		     exports    = 0,
		     getfiles   = 0,
                     invalid    = 0
		   }

run f s = snd (runState f s)

setLast stamp 
    = do (Stat f l tt su us se ex ge iv) <- get
         put (Stat f (max l stamp) tt su us se ex ge iv)
         return ()

setFirst stamp 
    = do (Stat f l tt su us se ex ge iv) <- get
         put (Stat (min f stamp) l tt su us se ex ge iv)
         return ()

incGetfiles  
    = do (Stat f l tt su us se ex ge iv) <- get
         put (Stat f l tt su us se ex (ge + 1) iv)
         return ()

incSearches  
    = do (Stat f l tt su us se ex ge iv) <- get
         put (Stat f l tt su us (se + 1) ex ge iv)
         return ()

incExports  
    = do (Stat f l tt su us se ex ge iv) <- get
         put (Stat f l tt su us se (ex + 1) ge iv)
         return ()

incTotal  
    = do (Stat f l tt su us se ex ge iv) <- get
         put (Stat f l (tt + 1) su us se ex ge iv)
         return ()

incSuccess  
    = do (Stat f l tt su us se ex ge iv) <- get
         put (Stat f l tt (su + 1) us se ex ge iv)
         return ()

incUser
    = do (Stat f l tt su us se ex ge iv) <- get
         put (Stat f l tt su (us + 1) se ex ge iv)
         return ()

incInvalid 
    = do (Stat f l tt su us se ex ge iv) <- get
         put (Stat f l tt su us se ex ge (iv + 1))
         return ()


isSuccess Ok        = True
isSuccess (Error _) = False

isUser User = True
isUser _    = False

isSearch (Search _) = True
isSearch _          = False

isExport (Export _) = True
isExport _          = False

isGetfile (Getfile _) = True
isGetfile _           = False



-- Actual update of the stats

_update (Just (Hit _ time query agent status))
    = incTotal 
      >> setFirst time 
      >> setLast time
      >> if (isSuccess status) then
         incSuccess 
         >> (
             if (isUser agent) then 
             incUser
             >> (if (isSearch query) then incSearches else return ())
             >> (if (isExport query) then incExports  else return ())
             >> (if (isGetfile query) then incGetfiles  else return ())
             else return ()
            )
         else return ()

_update _ = incInvalid


_testUpdate 
    = let time = toClockTime (CalendarTime { ctYear  = 2005, ctMonth = January, ctDay   = 18,
				             ctHour  = 13,   ctMin   = 0,       ctSec   = 38,
				             ctTZ    = 3600, ctPicosec = 0,
				             ctWDay = undefined, ctYDay = undefined, 
				             ctTZName = undefined, ctIsDST = False
				           })
          run' h = run (_update (Just h)) initialStat

      in HUnit.TestCase 
      ( do HUnit.assertEqual "invalid lines don't count"
                    (Stat {firstHit = future, lastHit = epoch,
                           total = 0, success = 0, user = 0, 
                           searches = 0, exports = 0, getfiles = 0, 
                           invalid = 1})
                    (run (_update Nothing) initialStat)

           HUnit.assertEqual "hit failure"
                    (Stat {firstHit = time, lastHit = time,
                           total = 1, success = 0, user = 0, 
                           searches = 0, exports = 0, getfiles = 0, 
                           invalid = 0})
                    (run' (Hit "" time (Browse "/") User (Error 404)) )

           HUnit.assertEqual "a user browse"
                    (Stat {firstHit = time, lastHit = time,
                           total = 1, success = 1, user = 1, 
                           searches = 0, exports = 0, getfiles = 0, 
                           invalid = 0})
                    (run' (Hit "" time (Browse "/") User Ok) )

           HUnit.assertEqual "a user search"
                    (Stat {firstHit = time, lastHit = time,
                           total = 1, success = 1, user = 1, 
                           searches = 1, exports = 0, getfiles = 0, 
                           invalid = 0})
                    (run' (Hit "" time (Search "") User Ok) )

           HUnit.assertEqual "a bot search should not be counted"
                    (Stat {firstHit = time, lastHit = time,
                           total = 1, success = 1, user = 0, 
                           searches = 0, exports = 0, getfiles = 0, 
                           invalid = 0})
                    (run' (Hit "" time (Search "") Bot Ok) )

           HUnit.assertEqual "a bot getfile should not be counted"
                    (Stat {firstHit = time, lastHit = time,
                           total = 1, success = 1, user = 0, 
                           searches = 0, exports = 0, getfiles = 0, 
                           invalid = 0})
                    (run' (Hit "" time (Getfile "") Bot Ok) )

           HUnit.assertEqual "a bot export should not be counted"
                    (Stat {firstHit = time, lastHit = time,
                           total = 1, success = 1, user = 0, 
                           searches = 0, exports = 0, getfiles = 0, 
                           invalid = 0})
                    (run' (Hit "" time (Export "") Bot Ok) )

           HUnit.assertEqual "a bot browse should not be counted"
                    (Stat {firstHit = time, lastHit = time,
                           total = 1, success = 1, user = 0, 
                           searches = 0, exports = 0, getfiles = 0, 
                           invalid = 0})
                    (run' (Hit "" time (Browse "") Bot Ok) )
      )


-- Public function to parse the logs

parseLog contents
    = let result  = foldl _parseLine initialStat $ lines contents
	  days    = 1 + (tdSec (lastHit result `diffClockTimes` firstHit result) `div` (3600 * 24))
	  mean    = total result `div` days
	  total'    = fromIntegral (total result)
	  success'  = fromIntegral (success result)
	  user'     = fromIntegral (user result)
	  searches' = fromIntegral (searches result)
	  exports'  = fromIntegral (exports result)
          avg val   = (show $ round (100 * val)) ++ "%"

	  in "begin: " ++ show (firstHit result) ++ "\n"
		 ++ "end:   " ++ show (lastHit result) ++ "\n"
		 ++ "total: " ++ show (total result) ++ " hits (" 
		 ++ show mean ++ " hits/day, " ++  avg (success' / total') ++ " success)\n"
		 ++ "users hits: " ++ avg (user' / total') ++ "\n"
		 ++ "searches:   " ++ avg (searches' / total') ++ "\n" 
		 ++ "exports:    " ++ avg (exports' / total') ++ "\n"
		 ++ "fulltexts:  " ++ show (getfiles result `div` days) ++ " per day\n"
                 ++ "\n(invalid records: " ++ (show (invalid result)) ++ ")\n"

    where _parseLine stat line = run (_update (parseHit line)) stat


testParseLog = HUnit.runTestTT (HUnit.TestList [_testUpdate])

